let verbosity = ref 2;;

type fs = Atom of string | DQ of string | Nil | Cons of fs * fs

let caseh : (string,string) Hashtbl.t = Hashtbl.create 100;;
let warnedcaseh : (string * string,unit) Hashtbl.t = Hashtbl.create 100;;
let caseissue = ref false;;
let vars = ref [];;

let var_p v =
  if String.length v > 0 then
    v.[0] = '?'
  else
    false

let rowvar_p rv =
  if String.length rv > 0 then
    rv.[0] = '@'
  else
    false

let var_or_rowvar_p v = var_p v || rowvar_p v

let clear_vars () =
  List.iter (fun v -> Hashtbl.remove caseh v) !vars;
  vars := []

let rec remember_case_and_upper x =
  let y = String.uppercase_ascii x in
  if var_or_rowvar_p y then vars := y::!vars;
  try
    let z = Hashtbl.find caseh y in
    if not (z = x) then
      begin
        if not (Hashtbl.mem warnedcaseh (x,z)) then
          begin
            Printf.printf "WARNING: %s previously used with case %s\n" x z;
            Hashtbl.add warnedcaseh (x,z) ()
          end;
        caseissue := true;
      end;
    y
  with Not_found ->
    Hashtbl.add caseh y x;
    y

let rec list_of_len_p l n =
  match l with
  | Nil -> n = 0
  | Cons(ca,cd) -> if n > 0 then list_of_len_p cd (n - 1) else false
  | _ -> false

let rec nonprim_use_of_prim a =
  match a with
  | Atom(x) when x = "INSTANCE" || x = "SUBCLASS" || x = "KAPPAFN" -> true
  | Cons(Atom(x),l) when x = "KAPPAFN" ->
     begin
       match l with
       | Cons(Atom(v),Cons(body,Nil)) -> not (var_p v) || nonprim_use_of_prim body
       | _ -> true
     end
  | Cons(Atom(x),l) when x = "INSTANCE" || x = "SUBCLASS" ->
     if list_of_len_p l 2 then
       nonprim_use_of_prim l
     else
       true
  | Cons(ca,cd) ->
     nonprim_use_of_prim ca || nonprim_use_of_prim cd
  | _ -> false

let rec rowvar_not_at_end_p a =
  match a with
  | Cons(Atom(q),Cons(l,body)) when q = "FORALL" || q = "EXISTS" -> rowvar_not_at_end_p body
  | Cons(Atom(rv),l) when rowvar_p rv -> not (l = Nil)
  | Cons(ca,cd) -> rowvar_not_at_end_p ca || rowvar_not_at_end_p cd
  | Atom(rv) when rowvar_p rv -> true
  | _ -> false

let bad_number_partition_p a =
  match a with
  | Cons(Atom(pa),Cons(Atom(nu),_)) when pa = "PARTITION" && nu = "NUMBER" -> true
  | _ -> false

let rec fs_str s =
  match s with
  | Atom(x) -> x
  | DQ(x) -> Printf.sprintf "\"%s\"" x
  | Nil -> "NIL"
  | Cons(s1,s2) -> Printf.sprintf "(%s%s)" (fs_str s1) (fs_rest_str s2)
and fs_rest_str s =
  match s with
  | Atom(x) -> Printf.sprintf " . %s" x
  | DQ(x) -> Printf.sprintf " . \"%s\"" x
  | Nil -> ""
  | Cons(s1,s2) -> Printf.sprintf " %s%s" (fs_str s1) (fs_rest_str s2)

let rec fs_str_dq s =
  match s with
  | Atom(x) -> Printf.sprintf "\"%s\"" x
  | DQ(x) -> Printf.sprintf "\"\\\"%s\\\"\"" x
  | Nil -> "NIL"
  | Cons(s1,s2) -> Printf.sprintf "(%s%s)" (fs_str_dq s1) (fs_rest_str_dq s2)
and fs_rest_str_dq s =
  match s with
  | Atom(x) -> Printf.sprintf " . \"%s\"" x
  | DQ(x) -> Printf.sprintf " . \"\\\"%s\\\"\"" x
  | Nil -> ""
  | Cons(s1,s2) -> Printf.sprintf " %s%s" (fs_str_dq s1) (fs_rest_str_dq s2)
	
let peek_char (oc,f) =
  match oc with
  | Some(c) -> (c,(oc,f))
  | None ->
      let c = input_char f in
      (c,(Some(c),f))

let pop_char (oc,f) =
  match oc with
  | Some(c) -> (c,(None,f))
  | None ->
      let c = input_char f in
      (c,(None,f))
    
let whitespace_p c = c = ' ' || c = '\n' || c = '\t' || c = '\r'

let rec consify_list l n =
  match l with
  | [] -> n
  | a::r -> Cons(a,consify_list r n)

let atom_of r =
  if r = "NIL" || r = "nil" || r = "Nil" then
    Nil
  else
    Atom(remember_case_and_upper r)

let rec parse_dquote f sofar =
  let (c,f) = pop_char f in
  if c = '"' then
    (sofar,f)
  else
    parse_dquote f (Printf.sprintf "%s%c" sofar c)

let rec parse_vbarname f sofar =
  let (c,f) = pop_char f in
  if c = '|' then
    (sofar,f)
  else
    parse_vbarname f (Printf.sprintf "%s%c" sofar c)

let rec parse_token r f =
  try
    let (c,f) = peek_char f in
    if c = '(' || c = ')' || c = ';' || c = '"' || c = '|' || whitespace_p c then
      (atom_of r,f)
    else
      let (c,f) = pop_char f in
      parse_token (Printf.sprintf "%s%c" r c) f
  with End_of_file -> (atom_of r,f)

let rec parse_fake_sexpr_r f s =
  try
    if !verbosity > 5 then (Printf.printf "parse_fake_sexpr_r in %d\n" (List.length s); List.iter (fun a -> Printf.printf ". %s\n" (fs_str a)) s; flush stdout);
    let (c,f) = pop_char f in
    if !verbosity > 5 then (Printf.printf "parse_fake_sexpr_r in %c\n" c; flush stdout);
    if c = '(' then
      let (a,f) = parse_fake_sexpr_r f [] in
      parse_fake_sexpr_r f (a::s)
    else if c = ')' then
      (consify_list (List.rev s) Nil,f)
    else if c = ';' then
      skip_to_newline f s
    else if c = '.' then
      let (a,f) = parse_fake_sexpr_1 f in
      let f = skip_to_rparen f in
      (consify_list (List.rev s) a,f)
    else if whitespace_p c then
      parse_fake_sexpr_r f s
    else if c = '|' then
      let (q,f) = parse_vbarname f "" in
      parse_fake_sexpr_r f (Atom(Printf.sprintf "|%s|" q)::s)
    else if c = '"' then
      let (q,f) = parse_dquote f "" in
      parse_fake_sexpr_r f (DQ(q)::s)
    else
      let (tok,f) = parse_token (Printf.sprintf "%c" c) f in
      parse_fake_sexpr_r f (tok::s)
  with End_of_file ->
    (consify_list (List.rev s) Nil,f)
and skip_to_newline f s =
  if !verbosity > 5 then (Printf.printf "skip_to_newline in\n"; flush stdout);
  let (c,f) = pop_char f in
  if !verbosity > 5 then (Printf.printf "skip_to_newline in %c\n" c; flush stdout);
  if c = '\n' then
    parse_fake_sexpr_r f s
  else
    skip_to_newline f s
and skip_to_rparen f =
  let (c,f) = pop_char f in
  if c = ')' then
    f
  else
    skip_to_rparen f
and skip_to_newline_1 f =
  if !verbosity > 5 then (Printf.printf "skip_to_newline_1 in\n"; flush stdout);
  let (c,f) = pop_char f in
  if !verbosity > 5 then (Printf.printf "skip_to_newline_1 in %c\n" c; flush stdout);
  if c = '\n' then
    parse_fake_sexpr_1 f
  else
    skip_to_newline_1 f
and parse_fake_sexpr_1 f =
  let (c,f) = pop_char f in
  if !verbosity > 5 then (Printf.printf "parse_fake_sexpr_1 in %c\n" c; flush stdout);
  if c = '(' then
    parse_fake_sexpr_r f []
  else if c = ')' then
    raise (Failure "unmatched )")
  else if c = ';' then
    skip_to_newline_1 f
  else if c = '.' then
    raise (Failure "unmatched .")
  else if whitespace_p c then
    parse_fake_sexpr_1 f
  else
    parse_token (Printf.sprintf "%c" c) f
      
let parse_fake_sexpr f =
  parse_fake_sexpr_1 (None,f)

let translate_fake_sexpr_file strfun fn gn =
  let f = open_in fn in
  let f2 = ref (None,f) in
  let g = open_out gn in
  try
    while true do
      caseissue := false;
      clear_vars ();
      let (a,s) = parse_fake_sexpr_1 !f2 in
      f2 := s;
      if !caseissue then
        begin
          if !verbosity > 1 then
            Printf.printf "Skipping entry due to case issue: %s\n" (strfun a)
        end
      else if nonprim_use_of_prim a then
        begin
          if !verbosity > 1 then
            Printf.printf "Skipping entry due to nonprimitive use of prim: %s\n" (strfun a)
        end
      else if rowvar_not_at_end_p a then
        begin
          if !verbosity > 1 then
            Printf.printf "Skipping entry due to row var not at end of list: %s\n" (strfun a)
        end
      else if bad_number_partition_p a then
        Printf.fprintf g "(SUBCLASS REALNUMBER COMPLEXNUMBER)\n(SUBCLASS COMPLEXNUMBER NUMBER)\n"
      else
        Printf.fprintf g "%s\n" (strfun a)
    done
  with
  | End_of_file ->
      close_in f;
      close_out g
  | exn ->
      close_in f;
      close_out g;
      raise exn

let translate_fake_sexpr_file_ndq fn gn =
  translate_fake_sexpr_file fs_str fn gn

let translate_fake_sexpr_file_dq fn gn =
  translate_fake_sexpr_file fs_str_dq fn gn

let rec translate_fake_sexpr_batch bfn =
  let bf = open_in bfn in
  try
    while true do
      let fn = input_line bf in
      let gn = input_line bf in
      translate_fake_sexpr_file fs_str fn gn
    done
  with End_of_file -> ()

let main () =
  let l = Array.length Sys.argv in
  if l < 3 || l mod 2 = 0 || (l > 3 && Sys.argv.(1) = "-batch") then
    begin
      Printf.printf "USAGE: sumo2asumo <in1> <out1> .. <ink> <outk>\n";
      Printf.printf "or sumo2asumo -batch <file>\nwhere <file> has alternating lines <in> and <out>.";
      exit 1;
    end;
  if Sys.argv.(1) = "-batch" then
    translate_fake_sexpr_batch Sys.argv.(2)
  else
    for i = 1 to (l - 1) / 2 do
      translate_fake_sexpr_file fs_str (Sys.argv.(2 * i - 1)) (Sys.argv.(2 * i));
    done;
  exit 0;;

main ()
  
